from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.tech import LAYER


@gf.cell
def pads_shorted(
    width: int = 100,
    n_pads: int = 8,
    pad_spacing: int = 150,
    layer: Tuple[int, int] = LAYER.M1,
) -> Component:
    c = Component(name="shorted_pads")
    pad = rectangle(size=(width, width), layer=layer, centered=True)
    for i in range(n_pads):
        pad_ref = c.add_ref(pad)
        pad_ref.movex(i * pad_spacing - n_pads / 2 * pad_spacing + pad_spacing / 2)

    short = rectangle(size=(pad_spacing * (n_pads - 1), 10), layer=layer, centered=True)
    c.add_ref(short)
    return c


if __name__ == "__main__":

    c = pads_shorted()
    c.show()



import pathlib
from typing import List, Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.tech import LAYER


def _add_pin_square(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.PORT,
    port_margin: float = 0.0,
) -> None:
    """Add half out pin to a component.

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label
        port_margin: margin to port edge


    .. code::

           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2 + port_margin

    dbot = np.array([pin_length / 2, -d])
    dtop = np.array([pin_length / 2, d])
    dbotin = np.array([-pin_length / 2, -d])
    dtopin = np.array([-pin_length / 2, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    if label_layer:
        component.add_label(
            text=str(p.name),
            position=p.midpoint,
            layer=label_layer,
        )


def _rotate(v, m):
    return np.dot(m, v)


def add_monitors_and_extend_ports(
    component: Component,
    port_labels: Optional[List[Port]] = None,
    extension_length: float = 0.2,
    port_margin: float = 1.0,
    layer: Tuple[int, int] = (101, 0),
) -> Component:
    """Add pins for port_labels
    each port increments the port layer

    Args:
        component: Component
        port_labels: list of labels
        extension_length: length to extend each port
        port_margin: margin from edge of the waveguide to edge of the port
        layer: monitor layer

    Returns: component with extended ports and monitors

    """

    port_labels = port_labels or component.ports.keys()

    for i, port_label in enumerate(port_labels):
        _add_pin_square(
            component=component,
            port=component.ports[port_label],
            pin_length=0,
            port_margin=port_margin,
            layer=(layer[0] + i, layer[1]),
        )

    return gf.components.extension.extend_ports(
        component=component, length=extension_length
    )


@cell
def add_monitors(
    component: Component,
    source_port_name: str = "o1",
    extension_length: float = 1.0,
    source_distance_to_monitors: float = 0.2,
    port_margin: float = 1.0,
    layer_monitor: Tuple[int, int] = (101, 0),
    layer_source: Tuple[int, int] = (110, 0),
    layer_simulation_region: Tuple[int, int] = (2, 0),
    top: float = 0.0,
    bottom: float = 0.0,
    right: float = 0.0,
    left: float = 0.0,
) -> Component:
    """Add monitors in layer_monitor, incrementing the layer number for each port
    Then extends all ports by source_distance_to_monitors
    Then add source in source_port_name
    Finally extends all ports by extension_length
    returns device centered at x=0, y=0

    Args:
        component: to add monitors
        source_port_name: name of the port to add the mode source
        extension_length: extension lenght when extending the ports
        port_margin: from waveguide edge to port edge
        layer_monitor: layer for the fist monitor
        layer_source: layer for the source
        layer_simulation_region: layer for the simulation region
        top: simulation region top padding
        bottom: simulation region south padding
        right: simulation region east padding
        left: simulation region west padding

    Returns: component with extended ports and monitors

    """
    c = gf.Component(f"{component.name}_monitors")

    # add monitors
    component_with_monitors = add_monitors_and_extend_ports(
        component=component,
        extension_length=source_distance_to_monitors,
        port_margin=port_margin,
        layer=layer_monitor,
    )

    # add source
    component_with_source = add_monitors_and_extend_ports(
        component_with_monitors,
        extension_length=extension_length,
        port_labels=[source_port_name],
        layer=layer_source,
    )

    # add simulation region
    component_with_padding = gf.add_padding(
        component=component_with_source,
        default=0,
        layers=[layer_simulation_region],
        top=top,
        bottom=bottom,
        right=right,
        left=left,
    )
    c.add(component_with_padding)
    c.ports = component_with_padding.ports
    return c


if __name__ == "__main__":
    gdspath = pathlib.Path.cwd() / "waveguide.gds"
    c = gf.components.bend_circular(radius=5)
    # c = gf.components.waveguide(length=2)
    # cm = extend_ports(component=c)
    cm = add_monitors(component=c)
    gf.show(cm)



from functools import partial

import meep as mp
import numpy as np

from gdsfactory.simulation.gmpb.find_modes import find_modes
from gdsfactory.simulation.gmpb.get_mode_solver_rib import get_mode_solver_rib
from gdsfactory.simulation.gmpb.types import WidthSweep


def find_neff_vs_width(
    w1: float = 0.2,
    w2: float = 1.0,
    steps: int = 12,
    mode_solver=get_mode_solver_rib,
    nmodes: int = 4,
    wavelength: float = 1.55,
    parity=mp.NO_PARITY,
    **kwargs
):
    width = np.linspace(w1, w2, steps)

    neff = {}

    for mode_number in range(1, nmodes + 1):
        neff[mode_number] = []

    for wg_width in width:
        mode_solver = partial(mode_solver, wg_width=wg_width, nmodes=nmodes, **kwargs)
        modes = find_modes(mode_solver, wavelength=wavelength, parity=parity)
        for mode_number in range(1, nmodes + 1):
            mode = modes[mode_number]
            neff[mode_number].append(mode.neff)

    return WidthSweep(width=list(width), neff=neff)


if __name__ == "__main__":
    # s = find_neff_vs_width()

    w1: float = 0.5
    w2: float = 1.0
    steps: int = 3
    mode_solver = get_mode_solver_rib
    nmodes: int = 4
    wavelength: float = 1.55
    parity = mp.NO_PARITY
    width = np.linspace(w1, w2, steps)

    neff = {}

    for mode_number in range(1, nmodes + 1):
        neff[mode_number] = []

    for wg_width in width:
        mode_solver = partial(mode_solver, wg_width=wg_width, nmodes=nmodes)
        modes = find_modes(mode_solver, wavelength=wavelength, parity=parity)
        for mode_number in range(1, nmodes + 1):
            mode = modes[mode_number]
            neff[mode_number].append(mode.neff)

    s = WidthSweep(width=list(width), neff=neff)



import gdsfactory as gf


def test_partial_function():
    mmi400 = gf.partial(gf.components.mmi1x2, width=0.4)
    mmi600 = gf.partial(gf.components.mmi1x2, width=0.6)
    mzi400 = gf.partial(gf.components.mzi, splitter=mmi400)
    mzi600 = gf.partial(gf.components.mzi, splitter=mmi600)
    c400 = mzi400()
    c600 = mzi600()
    assert c600.name != c400.name, print(c400.name, c600.name)


if __name__ == "__main__":
    mmi400 = gf.partial(gf.components.mmi1x2, width=0.4)
    mmi600 = gf.partial(gf.components.mmi1x2, width=0.6)
    mzi400 = gf.partial(gf.components.mzi, splitter=mmi400)
    mzi600 = gf.partial(gf.components.mzi, splitter=mmi600)
    c400 = mzi400()
    c600 = mzi600()
    print(c400.name, c600.name)



from typing import List, Optional, Tuple

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.tech import TECH

LAYER = TECH.layer
Layer = Tuple[int, int]


def get_padding_points(
    component: Component,
    default: float = 50.0,
    top: Optional[float] = None,
    bottom: Optional[float] = None,
    right: Optional[float] = None,
    left: Optional[float] = None,
) -> List[float]:
    """Returns padding points for a component outline.

    Args:
        component
        default: default padding
        top: north padding
        bottom: south padding
        right: east padding
        left: west padding
    """
    c = component
    top = top if top is not None else default
    bottom = bottom if bottom is not None else default
    right = right if right is not None else default
    left = left if left is not None else default
    return [
        [c.xmin - left, c.ymin - bottom],
        [c.xmax + right, c.ymin - bottom],
        [c.xmax + right, c.ymax + top],
        [c.xmin - left, c.ymax + top],
    ]


def add_padding(
    component: Component,
    layers: Tuple[Layer, ...] = (LAYER.PADDING,),
    **kwargs,
) -> Component:
    """Adds padding layers to a component.

    The cell name will be the same as the original component.

    Args:
        component
        layers: list of layers
        suffix for name
        default: default padding
        top: north padding
        bottom: south padding
        right: east padding
        left: west padding
    """
    points = get_padding_points(component, **kwargs)
    layers = layers or []
    for layer in layers:
        component.add_polygon(points, layer=layer)
    return component


@cell
def add_padding_container(
    component: Component,
    layers: Tuple[Layer, ...] = (LAYER.PADDING,),
    **kwargs,
) -> Component:
    """Adds padding layers to a component inside a container.

    Returns the same ports as the component.

    Args:
        component
        layers: list of layers
        default: default padding
        top: north padding
        bottom: south padding
        right: east padding
        left: west padding
    """

    c = Component()
    c.component = component
    cref = c << component

    points = get_padding_points(component, **kwargs)
    for layer in layers:
        c.add_polygon(points, layer=layer)
    c.ports = cref.ports
    c.copy_child_info(component)
    return c


def add_padding_to_size(
    component: Component,
    layers: Tuple[Layer, ...] = (LAYER.PADDING,),
    xsize: Optional[float] = None,
    ysize: Optional[float] = None,
    left: float = 0,
    bottom: float = 0,
) -> Component:
    """Returns component with padding layers on each side.

    New size is multiple of grid size
    """

    c = component
    top = abs(ysize - component.ysize) if ysize else 0
    right = abs(xsize - component.xsize) if xsize else 0
    points = [
        [c.xmin - left, c.ymin - bottom],
        [c.xmax + right, c.ymin - bottom],
        [c.xmax + right, c.ymax + top],
        [c.xmin - left, c.ymax + top],
    ]

    for layer in layers:
        component.add_polygon(points, layer=layer)

    return component


# def test_container():
#     import gdsfactory as gf
#     c = gf.components.straight(length=128)
#     cc = add_padding_container(component=c, layers=[(1, 0)])
#     assert len(cc.info["parent"]) == 5, len(cc.info['parent'])
#     cc = add_padding_container(component=c, layers=[(2, 0)])
#     assert len(cc.info["parent"]) == 5, len(cc.info['parent'])
#     cc = add_padding_container(component=c, layers=[(3, 0)])
#     assert len(cc.info["parent"]) == 5, len(cc.info['parent'])

# cc = add_padding_container(component=c, layers=[(4, 0)])
# assert isinstance(cc.settings["component"], Component)
# print(cc.settings["component"])
# print(len(cc.settings["component"]))


if __name__ == "__main__":
    # test_container()

    import gdsfactory as gf

    c = gf.components.straight(length=128)
    cc = add_padding_container(component=c, layers=[(2, 0)])
    print(cc.info["parent"])
    # cc.show()
    # cc.pprint()

    # c = gf.components.straight(length=5)
    # cc = add_padding_to_size(component=c, xsize=10, layers=[(2, 0)])
    # cc.show()
    # print(cc.settings)
    # print(cc.ports)



"""In programming, a factory is a function that returns an Object.

Functions are easy to understand because they have clear inputs and outputs.
Most gdsfactory functions take some inputs and return a Component object.
Some of these inputs parameters are also functions.

- Component: Object with.
    - name
    - references to other components (x, y, rotation)
    - polygons in different layers
    - ports dictionary
- ComponentFactory: function that returns a Component.
- Route: dataclass with 3 attributes.
    - references: list of references (straights, bends and tapers)
    - ports: dict(input=PortIn, output=PortOut)
    - length: float (how long is this route)
- RouteFactory: function that returns a Route.

"""
import dataclasses
import pathlib
from typing import Callable, Dict, List, Optional, Tuple, Union

from phidl.device_layout import Path

from gdsfactory.component import Component, ComponentReference
from gdsfactory.cross_section import CrossSection
from gdsfactory.port import Port


@dataclasses.dataclass
class Route:
    references: List[ComponentReference]
    ports: Tuple[Port, Port]
    length: float


@dataclasses.dataclass
class Routes:
    references: List[ComponentReference]
    lengths: List[float]
    ports: Optional[List[Port]] = None
    bend_radius: Optional[float] = None


Float2 = Tuple[float, float]
Float3 = Tuple[float, float, float]
Floats = Tuple[float, ...]
Strs = Tuple[str, ...]
Int2 = Tuple[int, int]
Int3 = Tuple[int, int, int]
Ints = Tuple[int, ...]

Layer = Tuple[int, int]
Layers = Tuple[Layer, ...]
RouteFactory = Callable[..., Route]
ComponentFactory = Callable[..., Component]
ComponentFactoryDict = Dict[str, ComponentFactory]
PathFactory = Callable[..., Path]
PathType = Union[str, pathlib.Path]
PathTypes = Tuple[PathType, ...]

ComponentOrFactory = Union[ComponentFactory, Component]
ComponentOrPath = Union[PathType, Component]
ComponentOrReference = Union[Component, ComponentReference]
NameToFunctionDict = Dict[str, ComponentFactory]
Number = Union[float, int]
Coordinate = Tuple[float, float]
Coordinates = Tuple[Coordinate, ...]
ComponentOrPath = Union[Component, PathType]
CrossSectionFactory = Callable[..., CrossSection]
CrossSectionOrFactory = Union[CrossSection, Callable[..., CrossSection]]


__all__ = (
    "ComponentFactory",
    "ComponentOrFactory",
    "ComponentOrPath",
    "ComponentOrReference",
    "Coordinate",
    "Coordinates",
    "CrossSectionFactory",
    "CrossSectionOrFactory",
    "Float2",
    "Float3",
    "Floats",
    "Int2",
    "Int3",
    "Ints",
    "Layer",
    "Layers",
    "NameToFunctionDict",
    "Number",
    "PathType",
    "PathTypes",
    "Route",
    "RouteFactory",
    "Routes",
    "Strs",
)



from typing import Optional, Tuple

import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.cross_section import strip
from gdsfactory.port import Port
from gdsfactory.types import CrossSectionFactory, Layer


@cell
def taper(
    length: float = 10.0,
    width1: float = 0.5,
    width2: Optional[float] = None,
    port: Optional[Port] = None,
    with_cladding_box: bool = True,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    """Linear taper.

    Args:
        length:
        width1: width of the west port
        width2: width of the east port
        port: can taper from a port instead of defining width1
        with_cladding_box: to avoid DRC acute angle errors in cladding
        cross_section:
        kwargs: cross_section settings

    .. plot::
      :include-source:

      import gdsfactory as gf

      c = gf.components.taper(width1=0.5, width2=5, length=3)
      c.plot()

    """
    x = cross_section(**kwargs)

    layers_cladding = x.info["layers_cladding"]
    layer = x.info["layer"]

    if isinstance(port, gf.Port) and width1 is None:
        width1 = port.width
    if width2 is None:
        width2 = width1

    y1 = width1 / 2
    y2 = width2 / 2

    xpts = [0, length, length, 0]
    ypts = [y1, y2, -y2, -y1]

    c = gf.Component()
    c.add_polygon((xpts, ypts), layer=layer)
    c.add_port(
        name="o1",
        midpoint=[0, 0],
        width=width1,
        orientation=180,
        layer=layer,
        cross_section=x,
    )
    c.add_port(
        name="o2",
        midpoint=[length, 0],
        width=width2,
        orientation=0,
        layer=layer,
        cross_section=x,
    )

    if with_cladding_box and x.info["layers_cladding"]:
        layers_cladding = x.info["layers_cladding"]
        cladding_offset = x.info["cladding_offset"]
        points = get_padding_points(
            component=c,
            default=0,
            bottom=cladding_offset,
            top=cladding_offset,
        )
        for layer in layers_cladding or []:
            c.add_polygon(points, layer=layer)

    c.info["length"] = length
    c.info["width1"] = float(width1)
    c.info["width2"] = float(width2)
    return c


@gf.cell
def taper_strip_to_ridge(
    length: float = 10.0,
    width1: float = 0.5,
    width2: float = 0.5,
    w_slab1: float = 0.15,
    w_slab2: float = 6.0,
    layer_wg: Layer = gf.LAYER.WG,
    layer_slab: Layer = gf.LAYER.SLAB90,
    with_slab_port: bool = False,
    layers_cladding: Optional[Tuple[Layer, ...]] = None,
    cladding_offset: float = 3.0,
) -> Component:
    r"""Linear taper from strip to rib

    Args:
        length:
        width1:
        width2:
        w_slab1
        w_slab2
        layer_wg:
        layer_slab:
        with_slab_port: adds a wide slab port
        layers_cladding
        cladding_offset:

    .. code::

                      __________________________
                     /           |
             _______/____________|______________
                   /             |
       width1     |w_slab1       | w_slab2  width2
             ______\_____________|______________
                    \            |
                     \__________________________

    """

    taper_wg = taper(length=length, width1=width1, width2=width2, layer=layer_wg)
    taper_slab = taper(length=length, width1=w_slab1, width2=w_slab2, layer=layer_slab)

    c = gf.Component()
    for _t in [taper_wg, taper_slab]:
        taper_ref = _t.ref()
        c.add(taper_ref)
        c.absorb(taper_ref)

    c.info["length"] = length
    c.add_port(name="o1", port=taper_wg.ports["o1"])
    if with_slab_port:
        c.add_port(name="o2", port=taper_slab.ports["o2"])
    else:
        c.add_port(name="o2", port=taper_wg.ports["o2"])

    if layers_cladding:
        points = get_padding_points(
            component=c,
            default=0,
            bottom=cladding_offset,
            top=cladding_offset,
        )
        for layer in layers_cladding:
            c.add_polygon(points, layer=layer)

    return c


@gf.cell
def taper_strip_to_ridge_trenches(
    length=10.0,
    width=0.5,
    slab_offset=3.0,
    trench_width=2.0,
    trench_layer=gf.LAYER.SLAB90,
    layer_wg=gf.LAYER.WG,
    trench_offset_after_wg=0.1,
):

    c = gf.Component()
    y0 = width / 2 + trench_width - trench_offset_after_wg
    yL = width / 2 + trench_width - trench_offset_after_wg + slab_offset

    # straight
    x = [0, length, length, 0]
    yw = [y0, yL, -yL, -y0]
    c.add_polygon((x, yw), layer=layer_wg)

    # top trench
    ymin0 = width / 2
    yminL = width / 2
    ymax0 = width / 2 + trench_width
    ymaxL = width / 2 + trench_width + slab_offset
    x = [0, length, length, 0]
    ytt = [ymin0, yminL, ymaxL, ymax0]
    ytb = [-ymin0, -yminL, -ymaxL, -ymax0]
    c.add_polygon((x, ytt), layer=trench_layer)
    c.add_polygon((x, ytb), layer=trench_layer)

    c.add_port(name="o1", midpoint=[0, 0], width=width, orientation=180, layer=layer_wg)
    c.add_port(
        name="o2", midpoint=[length, 0], width=width, orientation=0, layer=layer_wg
    )

    return c


if __name__ == "__main__":
    c = taper(width2=1)
    # c = taper_strip_to_ridge(with_slab_port=True, layers_cladding=((111, 0),))
    # print(c.get_optical_ports())
    # c = taper_strip_to_ridge_trenches()
    c.show()



from typing import Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.taper import taper


@gf.cell
def grating_coupler_rectangular(
    n_periods: int = 20,
    period: float = 0.75,
    fill_factor: float = 0.5,
    width_grating: float = 11.0,
    length_taper: float = 150.0,
    wg_width: float = 0.5,
    layer: Tuple[int, int] = gf.LAYER.WG,
    polarization: str = "te",
    wavelength: float = 1.55,
) -> Component:
    r"""Grating coupler uniform (grating with rectangular shape not elliptical).
    Therefore it needs a longer taper.
    Grating teeth are straight instead of elliptical.

    Args:
        n_periods: 20
        period: 0.75
        fill_factor: 0.5
        width_grating: 11
        length_taper: 150
        wg_width: 0.5

    .. code::

                 \  \  \  \
                  \  \  \  \
                _|-|_|-|_|-|___
               |_______________  W0

    """
    c = Component()
    taper_ref = c << taper(
        length=length_taper,
        width2=width_grating,
        width1=wg_width,
        layer=layer,
    )

    c.add_port(port=taper_ref.ports["o1"], name="o1")
    x0 = taper_ref.xmax

    for i in range(n_periods):
        xsize = gf.snap.snap_to_grid(period * fill_factor)
        cgrating = c.add_ref(
            rectangle(size=[xsize, width_grating], layer=layer, port_type=None)
        )
        cgrating.x = gf.snap.snap_to_grid(x0 + i * period)
        cgrating.y = 0

    xport = np.round((x0 + cgrating.x) / 2, 3)

    port_type = f"vertical_{polarization.lower()}"
    c.add_port(name=port_type, port_type=port_type, midpoint=(xport, 0), orientation=0)
    c.info.polarization = polarization
    c.info.wavelength = wavelength
    gf.asserts.grating_coupler(c)
    return c


if __name__ == "__main__":
    # c = grating_coupler_rectangular(name='gcu', partial_etch=True)
    c = grating_coupler_rectangular()
    print(c.ports)
    c.show()



import gdsfactory as gf
from gdsfactory.components.mzi import mzi
from gdsfactory.components.straight_heater_metal import (
    straight_heater_metal,
    straight_heater_metal_90_90,
)

mzi_phase_shifter = gf.partial(
    mzi, straight_x_top=straight_heater_metal, length_x=320.0
)

mzi_phase_shifter_90_90 = gf.partial(
    mzi_phase_shifter, straight_x_top=straight_heater_metal_90_90
)

if __name__ == "__main__":
    c = mzi_phase_shifter_90_90()
    c = mzi_phase_shifter(
        straight_x_top=gf.c.straight_pin, straight_x_bot=gf.c.straight_pin
    )
    c = mzi_phase_shifter(
        straight_x_top=gf.c.straight_heater_doped_rib,
        straight_x_bot=gf.c.straight_heater_doped_rib,
        delta_length=20,
    )
    c.show()
    print(c.name)



"""Define names, clean values for names.
"""
import functools
import hashlib
from typing import Any

import numpy as np
from phidl import Device, Port
from phidl.device_layout import Path as PathPhidl

from gdsfactory.hash_points import hash_points
from gdsfactory.snap import snap_to_grid

MAX_NAME_LENGTH = 32


def join_first_letters(name: str) -> str:
    """Join the first letter of a name separated with underscores.

    taper_length -> TL
    """
    return "".join([x[0] for x in name.split("_") if x])


# replace function_name prefix for some components
component_type_to_name = dict(phidl="phidl")


def get_component_name(component_type: str, *args, **kwargs) -> str:
    """Returns concatenated kwargs Key_Value."""
    name = component_type
    name += "_".join([clean_value(a) for a in args])
    for k, v in component_type_to_name.items():
        name = name.replace(k, v)
    if kwargs:
        name += "_" + dict2name(**kwargs)
    return name


def dict2hash(**kwargs) -> str:
    ignore_from_name = kwargs.pop("ignore_from_name", [])
    h = hashlib.sha256()
    for key in sorted(kwargs):
        if key not in ignore_from_name:
            value = kwargs[key]
            value = clean_value(value)
            h.update(f"{key}{value}".encode())
    return h.hexdigest()


def dict2name(prefix: str = "", **kwargs) -> str:
    """Returns name from a dict."""
    ignore_from_name = kwargs.pop("ignore_from_name", [])
    kv = []
    kwargs = kwargs.copy()
    kwargs.pop("layer_to_inclusion", "")

    for key in sorted(kwargs):
        if key not in ignore_from_name and isinstance(key, str):
            value = kwargs[key]
            # key = join_first_letters(key).upper()
            if value is not None:
                kv += [f"{key}{clean_value(value)}"]
    label = prefix + "_".join(kv)
    return clean_name(label)


def assert_first_letters_are_different(**kwargs):
    """Assert that the first letters for each key are different.

    Avoids name collisions of different args that start with the same first letter.
    """
    first_letters = [join_first_letters(k) for k in kwargs.keys()]
    if not len(set(first_letters)) == len(first_letters):
        raise ValueError(
            f"Possible name collision! {kwargs.keys()} repeats first letters {first_letters}",
            "you can separate your arguments with underscores",
            " (delta_length -> DL, delta_width -> DW",
        )


def print_first_letters_warning(**kwargs):
    """Prints kwargs that have same cell."""
    first_letters = [join_first_letters(k) for k in kwargs.keys()]
    if not len(set(first_letters)) == len(first_letters):
        print(
            f"Possible name collision! {kwargs.keys()} "
            f"repeats first letters {first_letters}"
            "you can separate your arguments with underscores"
            " (delta_length -> DL, delta_width -> DW"
        )


def clean_name(name: str) -> str:
    """Return a string with correct characters for a cell name.

    [a-zA-Z0-9]

    FIXME: only a few characters are currently replaced.
        This function has been updated only on case-by-case basis
    """
    replace_map = {
        " ": "_",
        "!": "",
        "?": "",
        "#": "_",
        "%": "_",
        "(": "",
        ")": "",
        "*": "_",
        ",": "_",
        "-": "m",
        ".": "p",
        "/": "_",
        ":": "_",
        "=": "",
        "@": "_",
        "[": "",
        "]": "",
    }
    for k, v in list(replace_map.items()):
        name = name.replace(k, v)
    return name


def clean_value(value: Any) -> str:
    """returns more readable value (integer)
    if number is < 1:
        returns number units in nm (integer)

    units are in um by default. Therefore when we multiply by 1e3 we get nm.
    """

    if isinstance(value, int):
        value = str(value)
    elif isinstance(value, (float, np.float64)):
        if 1 > value > 1e-3:
            value = f"{int(value*1e3)}n"
        # elif float(int(value)) == value:
        #     value = str(int(value))
        elif 1e-6 < value < 1e-3:
            value = f"{snap_to_grid(value*1e6)}u"
        elif 1e-9 < value < 1e-6:
            value = f"{snap_to_grid(value*1e9)}n"
        elif 1e-12 < value < 1e-9:
            value = f"{snap_to_grid(value*1e12)}p"
        else:
            value = str(snap_to_grid(value)).replace(".", "p")
    elif isinstance(value, list):
        value = "_".join(clean_value(v) for v in value)
    elif isinstance(value, tuple):
        value = "_".join(clean_value(v) for v in value)
    elif isinstance(value, dict):
        value = dict2name(**value)
        # value = [f"{k}={v!r}" for k, v in value.items()]
    elif isinstance(value, Device):
        value = clean_name(value.name)
    elif isinstance(value, Port):
        value = f"{value.name}_{value.width}_{value.x}_{value.y}"
    elif isinstance(value, PathPhidl):
        value = f"path_{hash_points(value.points)}"
    elif (
        isinstance(value, object)
        and hasattr(value, "name")
        and isinstance(value.name, str)
    ):
        value = clean_name(value.name)
    elif callable(value) and hasattr(value, "__name__"):
        value = value.__name__
    elif callable(value) and isinstance(value, functools.partial):
        value = value.func.__name__ + dict2name(**value.keywords)
    elif isinstance(value, str):
        value = value.strip()
    elif hasattr(value, "get_name"):
        value = value.get_name()
    return str(value)


def get_name_short(name: str):
    if len(name) > MAX_NAME_LENGTH:
        name_hash = hashlib.md5(name.encode()).hexdigest()[:8]
        name = f"{name[:(MAX_NAME_LENGTH - 9)]}_{name_hash}"
    return clean_name(name)


def get_name(name: str) -> str:
    """Returns name with correct number of characters"""
    if not isinstance(name, str):
        raise ValueError(f"{name} needs to be a string")
    if len(name) > MAX_NAME_LENGTH:
        name_hash = hashlib.md5(name.encode()).hexdigest()[:8]
        name = f"{name[:(MAX_NAME_LENGTH - 9)]}_{name_hash}"
    return clean_name(name)


def test_clean_value() -> None:
    assert clean_value(0.5) == "500n"
    assert clean_value(5) == "5"
    assert clean_value(5.0) == "5p0"
    assert clean_value(11.001) == "11p001"


def test_clean_name() -> None:
    assert clean_name("wg(:_=_2852") == "wg___2852"


if __name__ == "__main__":
    # test_cell()
    test_clean_value()
    import gdsfactory as gf

    # print(clean_value(gf.components.straight))
    # c = gf.components.straight(polarization="TMeraer")
    # print(c.settings["polarization"])
    # print(clean_value(11.001))
    # layers_cladding = (gf.LAYER.WGCLAD, gf.LAYER.NO_TILE_SI)
    # layers_cladding = (gf.LAYER.WGCLAD,)
    c = gf.components.straight(length=10)
    c = gf.components.straight(length=10)

    # print(c.name)
    # print(c)
    # c.show()

    # print(clean_name("Waveguidenol1_(:_=_2852"))
    # print(clean_value(1.2))
    # print(clean_value(0.2))
    # print(clean_value([1, [2.4324324, 3]]))
    # print(clean_value([1, 2.4324324, 3]))
    # print(clean_value((0.001, 24)))
    # print(clean_value({"a": 1, "b": 2}))



"""FIXME! add sidewall angles."""
import pathlib
import tempfile
from typing import Tuple

import meep as mp
import numpy as np
import pydantic
from meep import mpb

mpb.Verbosity(0)

tmp = pathlib.Path(tempfile.TemporaryDirectory().name).parent / "meep"
tmp.mkdir(exist_ok=True)

Floats = Tuple[float, ...]


@pydantic.validate_arguments
def get_mode_solver_coupler(
    wg_widths: Floats = (0.5, 0.5),
    gaps: Floats = (0.2,),
    wg_thickness: float = 0.22,
    slab_thickness: int = 0.0,
    ncore: float = 3.47,
    nclad: float = 1.44,
    ymargin: float = 2.0,
    sz: float = 2.0,
    res: int = 32,
    nmodes: int = 4,
) -> mpb.ModeSolver:
    """Returns a mode_solver simulation.

    Args:
        wg_width: wg_width (um)
        wg_thickness: wg height (um)
        slab_thickness: thickness for the waveguide slab
        ncore: core material refractive index
        nclad: clad material refractive index
        sy: simulation region width (um)
        sz: simulation region thickness (um)
        res: resolution (pixels/um)
        nmodes: number of modes

    ::

          _____________________________________________________
          |
          |
          |         widths[0]                 widths[1]
          |     <---------->     gaps[0]    <---------->
          |      ___________ <------------->  ___________     _
          |     |           |               |           |     |
        sz|_____|           |_______________|           |_____|
          |                                                   | wg_thickness
          |slab_thickness                                     |
          |___________________________________________________|
          |
          |<--->                                         <--->
          |ymargin                                       ymargin
          |____________________________________________________
          <--------------------------------------------------->
                                   sy



    """
    material_core = mp.Medium(index=ncore)
    material_clad = mp.Medium(index=nclad)

    # Define the computational cell.  We'll make x the propagation direction.
    # the other cell sizes should be big enough so that the boundaries are
    # far away from the mode field.

    sy = np.sum(wg_widths) + np.sum(gaps) + 2 * ymargin
    geometry_lattice = mp.Lattice(size=mp.Vector3(0, sy, sz))

    # define the 2D blocks for the strip and substrate
    geometry = [
        mp.Block(
            size=mp.Vector3(mp.inf, mp.inf, mp.inf),
            material=material_clad,
        ),
        mp.Block(
            size=mp.Vector3(mp.inf, mp.inf, slab_thickness),
            material=material_core,
            center=mp.Vector3(z=-0.5 * slab_thickness),
        ),
    ]

    y = -sy / 2 + ymargin

    gaps = list(gaps) + [0]
    for i, wg_width in enumerate(wg_widths):
        geometry.append(
            mp.Block(
                size=mp.Vector3(mp.inf, wg_width, wg_thickness),
                material=material_core,
                center=mp.Vector3(y=y + wg_width / 2, z=0),
            )
        )

        y += gaps[i] + wg_width

    # The k (i.e. beta, i.e. propagation constant) points to look at, in
    # units of 2*pi/um.  We'll look at num_k points from k_min to k_max.
    num_k = 9
    k_min = 0.1
    k_max = 3.0
    k_points = mp.interpolate(num_k, [mp.Vector3(k_min), mp.Vector3(k_max)])

    # Increase this to see more modes.  (The guided ones are the ones below the
    # light line, i.e. those with frequencies < kmag / 1.45, where kmag
    # is the corresponding column in the output if you grep for "freqs:".)
    # use this prefix for output files

    wg_widths_str = "_".join([str(i) for i in wg_widths])
    gaps_str = "_".join([str(i) for i in gaps])
    filename_prefix = (
        tmp / f"coupler_{wg_widths_str}_{gaps_str}_{wg_thickness}_{slab_thickness}"
    )

    mode_solver = mpb.ModeSolver(
        geometry_lattice=geometry_lattice,
        geometry=geometry,
        k_points=k_points,
        resolution=res,
        num_bands=nmodes,
        filename_prefix=str(filename_prefix),
    )
    mode_solver.nmodes = nmodes
    mode_solver.info = dict(
        wg_widths=wg_widths,
        gaps=gaps,
        wg_thickness=wg_thickness,
        slab_thickness=slab_thickness,
        ncore=ncore,
        nclad=nclad,
        sy=sy,
        sz=sz,
        res=res,
        nmodes=nmodes,
    )
    return mode_solver


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    m = get_mode_solver_coupler()
    # m.init_params(p=0.5, reset_fields=False)
    eps = m.get_epsilon()
    cmap = "viridis"
    origin = "upper"
    plt.imshow(
        eps ** 0.5,
        cmap=cmap,
        origin=origin,
        aspect="auto",
    )



from gdsfactory.simulation.gmpb.find_mode_dispersion import find_mode_dispersion
from gdsfactory.simulation.gmpb.find_modes import find_modes
from gdsfactory.simulation.gmpb.find_neff_vs_width import find_neff_vs_width
from gdsfactory.simulation.gmpb.get_mode_solver_coupler import get_mode_solver_coupler
from gdsfactory.simulation.gmpb.get_mode_solver_rib import get_mode_solver_rib

__all__ = [
    "find_modes",
    "find_neff_vs_width",
    "find_mode_dispersion",
    "get_mode_solver_rib",
    "get_mode_solver_coupler",
]
__version__ = "0.0.2"



from typing import Optional, Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.contact import contact_m1_m3 as contact_metal_function
from gdsfactory.components.contact import contact_slab_npp_m3
from gdsfactory.components.taper_cross_section import taper_cross_section
from gdsfactory.cross_section import rib_heater_doped, strip_rib_tip
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import ComponentFactory, CrossSectionFactory, Layer


@gf.cell
def straight_heater_doped_rib(
    length: float = 320.0,
    nsections: int = 3,
    cross_section: CrossSectionFactory = strip_rib_tip,
    cross_section_heater: CrossSectionFactory = rib_heater_doped,
    contact_contact: Optional[ComponentFactory] = contact_slab_npp_m3,
    contact_metal: Optional[ComponentFactory] = contact_metal_function,
    contact_metal_size: Tuple[float, float] = (10.0, 10.0),
    contact_contact_size: Tuple[float, float] = (10.0, 10.0),
    contact_contact_yspacing: float = 2.0,
    taper: Optional[ComponentFactory] = taper_cross_section,
    taper_length: float = 10.0,
    **kwargs,
) -> Component:
    r"""Returns a doped thermal phase shifter.
    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        length: of the waveguide
        nsections: between contacts
        cross_section_heater: for the heater
        contact_contact: function to connect the heated strip
        contact_metal: function to connect the metal area
        contact_metal_size:
        contact_contact_yspacing: spacing from waveguide to contact
        kwargs: cross_section settings
        taper: optional taper
        taper_length:

    .. code::

                              length
          <-------------------------------------------->
                       length_section
             <--------------------------->
           length_contact
             <------->                             taper
             ______________________________________
           /|        |____________________|        |\
        __/ |viastack|                    |        | \___
        __  |        |                    |        |  ___cross_section
          \ | size   |____________________|        | /
           \|________|____________________|________|/

        taper         cross_section_heater



                                   |<------width------>|
                                    ____________________ heater_gap             slab_gap
                                   |                   |<---------->|               <-->
         ___ ______________________|                   |___________________________|___
        |   |            |               undoped Si                 |              |   |
        |   |layer_heater|               intrinsic region           |layer_heater  |   |
        |___|____________|__________________________________________|______________|___|
                                                                     <------------>
                                                                      heater_width
        <------------------------------------------------------------------------------>
                                       slab_width

    """
    c = Component()

    if taper:
        taper = taper(
            cross_section1=cross_section,
            cross_section2=cross_section_heater,
            length=taper_length,
            **kwargs,
        )
        length -= taper_length * 2

    wg = c << gf.c.straight(
        cross_section=cross_section_heater,
        length=snap_to_grid(length),
        **kwargs,
    )

    if taper:
        taper1 = c << taper
        taper2 = c << taper
        taper1.connect("o2", wg.ports["o1"])
        taper2.connect("o2", wg.ports["o2"])
        c.add_port("o1", port=taper1.ports["o1"])
        c.add_port("o2", port=taper2.ports["o1"])

    else:
        c.add_port("o1", port=wg.ports["o1"])
        c.add_port("o2", port=wg.ports["o2"])

    if contact_metal:
        contact_section = contact_metal(size=contact_metal_size)
    contacts = []
    length_contact = snap_to_grid(contact_contact_size[1])
    length_section = snap_to_grid((length - length_contact) / nsections)
    x0 = contact_contact_size[0] / 2
    for i in range(0, nsections + 1):
        xi = x0 + length_section * i

        if contact_metal and contact_contact:
            contact_center = c.add_ref(contact_section)
            contact_center.x = xi
            contact = c << contact_section
            contact.x = xi
            contact.y = +contact_metal_size[1] if i % 2 == 0 else -contact_metal_size[1]
            contacts.append(contact)

        if contact_contact:
            contact_bot = c << contact_contact(size=contact_contact_size)
            contact_top = c << contact_contact(size=contact_contact_size)
            contact_top.x = xi
            contact_bot.x = xi
            contact_top.ymin = +contact_contact_yspacing
            contact_bot.ymax = -contact_contact_yspacing

    if contact_metal and contact_contact:
        contact_length = length + contact_metal_size[0]
        contact_top = c << contact_metal(
            size=(contact_length, contact_metal_size[0]),
        )
        contact_bot = c << contact_metal(
            size=(contact_length, contact_metal_size[0]),
        )

        contact_bot.xmin = contacts[0].xmin
        contact_top.xmin = contacts[0].xmin

        contact_top.ymin = contacts[0].ymax
        contact_bot.ymax = contacts[1].ymin

        c.add_ports(contact_top.ports, prefix="top_")
        c.add_ports(contact_bot.ports, prefix="bot_")
    return c


def test_straight_heater_doped_rib_ports() -> Component:
    c = straight_heater_doped_rib(length=100.0)
    assert c.get_ports_xsize(port_type="optical") == 100.0, c.get_ports_xsize(
        port_type="optical"
    )
    return c


@gf.cell
def straight_heater_doped_rib_south(
    contact_metal_size: Tuple[float, float] = (10.0, 10.0),
    layer_metal: Layer = (49, 0),
    **kwargs,
):
    c = gf.Component()
    s = c << straight_heater_doped_rib()
    contact_metal_size = (10.0, 10.0)
    lbend = c << gf.c.L(
        width=contact_metal_size[0], size=(10, s.ysize), layer=layer_metal
    )
    lbend.connect("e2", s.ports["top_e3"])
    c.add_ports(s.ports)
    c.ports.pop("top_e3")
    c.add_port("e1", port=lbend.ports["e1"])
    return c


if __name__ == "__main__":
    # c = straight_heater_doped_rib(length=80, contact_metal=None, contact_contact=None)
    c = straight_heater_doped_rib(length=80)
    # c = test_straight_heater_doped_rib_ports()

    # c = gf.Component()
    # s = c << straight_heater_doped_rib()
    # contact_metal_size = (10.0, 10.0)
    # lbend = c << gf.c.L(width=contact_metal_size[0], size=(10, s.ysize))
    # lbend.connect("o2", s.ports["e1"])
    # c = straight_heater_doped_rib_south()
    c.show()



"""This example shows how to add_grating couplers

for single fiber INPUT single fiber OUTPU (no fiber array).
"""

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.samples.big_device import big_device


def test_fiber_single() -> Component:
    w = h = 18 * 50
    c = big_device(port_pitch=50.0, h=h, w=w)
    return gf.routing.add_fiber_single(component=c, zero_port="W1")


if __name__ == "__main__":
    c = test_fiber_single()
    c.show()



from typing import Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.types import ComponentFactory, Floats, Layer

_gaps = (0.2,) * 10
_widths = (0.5,) * 10


@gf.cell
def grating_coupler_rectangular_arbitrary(
    gaps: Floats = _gaps,
    widths: Floats = _widths,
    wg_width: float = 0.5,
    width_grating: float = 11.0,
    length_taper: float = 150.0,
    layer: Tuple[int, int] = gf.LAYER.WG,
    polarization: str = "te",
    wavelength: float = 1.55,
    taper: ComponentFactory = taper_function,
    layer_grating: Optional[Layer] = None,
    width_teeth: Optional[float] = None,
) -> Component:
    r"""Grating coupler uniform (grating with rectangular shape not elliptical).
    Therefore it needs a longer taper.
    Grating teeth are straight instead of elliptical.

    Args:
        gaps: list of gaps
        widths: list of widths
        wg_width:
        width_grating:
        length_taper:
        layer: for grating
        polarization:
        wavelength:
        taper: function
        layer_grating:
        width_teeth: defautls to width_grating

    .. code::

                 \  \  \  \
                  \  \  \  \
                _|-|_|-|_|-|___
               |_______________  W0

    """
    c = Component()
    taper_ref = c << taper(
        length=length_taper,
        width2=width_grating,
        width1=wg_width,
        layer=layer,
    )

    c.add_port(port=taper_ref.ports["o1"], name="o1")
    xi = taper_ref.xmax

    widths = gf.snap.snap_to_grid(widths)
    gaps = gf.snap.snap_to_grid(gaps)

    for width, gap in zip(widths, gaps):
        xi += gap + width / 2
        cgrating = c.add_ref(
            rectangle(
                size=[width, width_teeth or width_grating],
                layer=layer,
                port_type=None,
                centered=True,
            )
        )
        cgrating.x = gf.snap.snap_to_grid(xi)
        cgrating.y = 0
        xi += width / 2

    xport = np.round((xi + length_taper) / 2, 3)

    port_type = f"vertical_{polarization.lower()}"
    c.add_port(name=port_type, port_type=port_type, midpoint=(xport, 0), orientation=0)
    c.info.polarization = polarization
    c.info.wavelength = wavelength
    gf.asserts.grating_coupler(c)
    return c


if __name__ == "__main__":
    c = grating_coupler_rectangular_arbitrary()
    print(c.ports)
    c.show()



import gdsfactory as gf
from gdsfactory.import_gds import import_gds


def test_import_gds_snap_to_grid() -> None:
    gdspath = gf.CONFIG["gdsdir"] / "mmi1x2.gds"
    c = import_gds(gdspath, snap_to_grid_nm=5)
    assert len(c.get_polygons()) == 8, len(c.get_polygons())

    for x, y in c.get_polygons()[0]:
        assert gf.snap.is_on_grid(x, 5)
        assert gf.snap.is_on_grid(y, 5)


def test_import_gds_hierarchy() -> None:
    c0 = gf.components.mzi()
    gdspath = c0.write_gds()
    c = import_gds(gdspath)
    assert len(c.get_dependencies()) == 3, len(c.get_dependencies())


if __name__ == "__main__":
    # test_import_gds_hierarchy()
    test_import_gds_snap_to_grid()



"""# Components.

We store our component functions inside the `gf.components` module. Each function there returns a Component object

You can use `dir` the `gf.components` module to see the all available components.

Some of which are just shapes, but we call them components as they all inherit from the component class in `gf.Component`
"""


import gdsfactory as gf


def test_components_module() -> None:
    assert len(dir(gf.components)) > 1


if __name__ == "__main__":
    print(dir(gf.components))
    print(len(dir(gf.components)))



"""
INFO_VERSION

1: original metadata format

"""
import functools
import hashlib
import inspect
from typing import Dict

import omegaconf
from pydantic import validate_arguments

from gdsfactory.component import Component, clean_dict
from gdsfactory.name import MAX_NAME_LENGTH, clean_name, clean_value

CACHE: Dict[str, Component] = {}
INFO_VERSION = 1


class CellReturnTypeError(ValueError):
    pass


def clear_cache() -> None:
    """Clears the cache of components."""
    global CACHE
    CACHE = {}


def print_cache():
    for k in CACHE:
        print(k)


def clean_doc(name: str) -> str:
    """Returns a clean docstring"""
    # replace_map = {
    #     " ": " ",
    #     "  ": " ",
    #     "   ": " ",
    #     "    ": ",",
    #     "     ": " ",
    #     "\n": " ",
    #     "\n\n": " ",
    # }
    # for k, v in list(replace_map.items()):
    #     name = name.replace(k, v)

    # name = ",".join(name.split('\n'))
    # name = " ".join(name.split())
    return name


def cell_without_validator(func):
    @functools.wraps(func)
    def _cell(*args, **kwargs):
        """Decorator for Component functions.

        similar to cell decorator, this one does not validate_arguments using
        type annotations

        kwargs:
            name (str): Optional (ignored when autoname=True)
            cache (bool): get component from the cache if it already exists.
              Useful in jupyter notebook, so you don't have to clear the cache
            info: updates component.info dict
            prefix: name_prefix, defaults to function name
            max_name_length: truncates name beyond some characters (32) with a hash
            decorator: function to run over the component

        Implements a cache so that if a component has already been build
        it will return the component from the cache.
        This avoids 2 exact cells that are not references of the same cell
        You can always over-ride this with `cache = False`.

        .. plot::
          :include-source:

          import gdsfactory as gf

          @gf.cell
          def rectangle(size=(4,2), layer=0)->gf.Component:
              c = gf.Component()
              w, h = size
              points = [[w, h], [w, 0], [0, 0], [0, h]]
              c.add_polygon(points, layer=layer)
              return c

          c = rectangle(layer=(1,0))
          c.plot()

        """
        name = kwargs.pop("name", None)
        cache = kwargs.pop("cache", True)
        info = kwargs.pop("info", omegaconf.DictConfig({}))
        prefix = kwargs.pop("prefix", func.__name__)
        max_name_length = kwargs.pop("max_name_length", MAX_NAME_LENGTH)

        sig = inspect.signature(func)
        args_as_kwargs = dict(zip(sig.parameters.keys(), args))
        args_as_kwargs.update(**kwargs)
        args_as_kwargs_string_list = [
            f"{key}={clean_value(args_as_kwargs[key])}"
            for key in sorted(args_as_kwargs.keys())
        ]

        # for key in sorted(args_as_kwargs.keys()):
        #     print(f"{key}={clean_value(args_as_kwargs[key])}")

        arguments = "_".join(args_as_kwargs_string_list)
        arguments_hash = hashlib.md5(arguments.encode()).hexdigest()[:8]

        name_signature = (
            clean_name(f"{prefix}_{arguments_hash}") if arguments else prefix
        )
        name = name or name_signature
        decorator = kwargs.pop("decorator", None)

        if len(name) > max_name_length:
            name_hash = hashlib.md5(name.encode()).hexdigest()[:8]
            name = f"{name[:(MAX_NAME_LENGTH - 9)]}_{name_hash}"

        if (
            "args" not in sig.parameters
            and "kwargs" not in sig.parameters
            and "settings" not in sig.parameters
        ):
            for key in kwargs.keys():
                if key not in sig.parameters.keys():
                    raise TypeError(
                        f"{func.__name__}() got invalid argument `{key}`\n"
                        f"valid arguments are {list(sig.parameters.keys())}"
                    )

        if cache and name in CACHE:
            # print(f"CACHE LOAD {name} {func.__name__}({arguments})")
            return CACHE[name]
        else:
            # print(f"BUILD {name} {func.__name__}({arguments})")
            assert callable(
                func
            ), f"{func} got decorated with @cell! @cell decorator is only for functions"

            component = func(*args, **kwargs)

            if decorator:
                assert callable(
                    decorator
                ), f"decorator = {type(decorator)} needs to be callable"
                component_new = decorator(component)
                component = component_new or component

            if not isinstance(component, Component):
                raise CellReturnTypeError(
                    f"function `{func.__name__}` return type = `{type(component)}`",
                    "make sure that functions with @cell decorator return a Component",
                )

            component.name = name
            component.info.name = name
            component.info.module = func.__module__
            component.info.function_name = func.__name__
            component.info.info_version = INFO_VERSION

            default = {
                p.name: p.default
                for p in sig.parameters.values()
                if not callable(p.default)
            }
            full = default.copy()
            full.update(**args_as_kwargs)
            changed = args_as_kwargs.copy()

            clean_dict(full)
            clean_dict(default)
            clean_dict(changed)

            component.info.changed = changed
            component.info.default = default
            component.info.full = full

            component.info.update(**info)
            CACHE[name] = component
            return component

    return _cell


def cell(func, *args, **kwargs):
    """Validates type annotations with pydantic"""
    return cell_without_validator(validate_arguments(func), *args, **kwargs)


@cell
def wg(length: int = 3, width: float = 0.5) -> Component:
    """Dummy component for testing."""
    from gdsfactory.component import Component

    c = Component("straight")
    w = width / 2
    layer = (1, 0)
    c.add_polygon([(0, -w), (length, -w), (length, w), (0, w)], layer=layer)
    c.add_port(name="o1", midpoint=[0, 0], width=width, orientation=180, layer=layer)
    c.add_port(name="o2", midpoint=[length, 0], width=width, orientation=0, layer=layer)
    return c


def test_autoname() -> None:
    c = wg(length=3)
    # assert c.name == "wg_length3", c.name
    assert c.name == "wg_2dcab9f2", c.name


def test_set_name() -> None:
    c = wg(length=3, name="hi_there")
    assert c.name == "hi_there", c.name


@cell
def _dummy(length: int = 3, wg_width: float = 0.5) -> Component:
    """Dummy cell"""
    c = Component()
    w = length
    h = wg_width
    points = [
        [-w / 2.0, -h / 2.0],
        [-w / 2.0, h / 2],
        [w / 2, h / 2],
        [w / 2, -h / 2.0],
    ]
    c.add_polygon(points)
    return c


def test_names() -> None:
    name_base = _dummy().name
    assert name_base == "_dummy", name_base

    name_int = _dummy(length=3).name
    assert name_int == "_dummy_2dcab9f2", name_int

    dummy2 = functools.partial(_dummy, length=3)
    component_int = dummy2(length=3)
    name_int = component_int.name
    assert name_int == "_dummy_2dcab9f2", name_int
    # assert component_int.info.doc == "Dummy cell"

    name_float = _dummy(wg_width=0.5).name
    assert name_float == "_dummy_b78ec006", name_float

    name_length_first = _dummy(length=3, wg_width=0.5).name
    name_width_first = _dummy(wg_width=0.5, length=3).name
    assert (
        name_length_first == name_width_first
    ), f"{name_length_first} != {name_width_first}"

    name_with_prefix = _dummy(prefix="hi").name
    assert name_with_prefix == "hi", name_with_prefix

    name_args = _dummy(3).name
    name_kwargs = _dummy(length=3).name
    assert name_args == name_kwargs, name_with_prefix


if __name__ == "__main__":
    c = wg(name="my_waveguide")
    print(c.name)

    # dummy2 = functools.partial(_dummy, length=3)
    # c = dummy2()

    # c = _dummy()
    # test_raise_error_args()
    # c = gf.components.straight()

    # test_autoname_false()
    # test_autoname_true()
    # test_autoname()
    # test_set_name()

    # c = wg(length=3)
    # c = wg(length=3, autoname=False)

    # import gdsfactory as gf
    # info = dict(polarization="te")

    # c = gf.components.straight()
    # c = gf.components.straight(info=info)
    # c = gf.components.straight(length=3, info=info)
    # print(c.info.polarization)

    # print(c.settings.info.doc)
    # c = gf.components.spiral_inner_io(length=1e3)
    # c = gf.components.straight(length=3)
    # print(c.name)
    # c.show()

    # D = gf.Component()
    # arc = D << gf.components.bend_circular(
    #     radius=10, width=0.5, angle=90, layer=(1, 0), info=dict(polarization="te")
    # )
    # arc.rotate(90)
    # rect = D << gf.components.bbox(bbox=arc.bbox, layer=(0, 0))
